"""
Interactive
"""
import sys
import unicodedata
import argparse
import jumany

def column_print(data: []):
    """ Print data in columns   """
    columns = [12, 12, 12, 10, 10, 12, 10]
    for (show, col_width) in zip(data, columns):
        width = sum([
            2 if unicodedata.east_asian_width(x) in "FWA" else 1 for x in show
        ])
        while width > col_width:
            width -= 2 if unicodedata.east_asian_width(show[len(show)-1]) in "FWA" else 1
            show = show[:len(show)-1]
        if width < col_width:
            show += (' ' * (col_width - width))
        sys.stdout.write(show)
    sys.stdout.write("\n")

parser = argparse.ArgumentParser(
    description="""Results are shown in columns by default.
    Specify delimiter not to show in columns.
    Type 'exit' in half-width charactors to exit."""
)
parser.add_argument(
    '-d', dest='delimiter', nargs='?', const=' ',
    help='delimiter of parameters. default is a space'
)
parser.add_argument(
    '-r', dest='rcf', help='path to resource file'
)
args = parser.parse_args()

if jumany.open_lib(args.rcf) is False:
    print(jumany.get_error_msg())
    sys.exit(0)

while True:
    line = sys.stdin.readline()
    if line.strip() == "exit":
        break
    Mrphs = jumany.analyze(line, True)
    if Mrphs is None:
        print(jumany.get_error_msg())
        continue
    for (Midasi1, Yomi, Midasi2, Hinsi, Bunrui, Katuyou1, Katuyou2) in Mrphs:
        mrph = [
            Midasi1, Yomi, Midasi2, jumany.get_hinsi(Hinsi),
            jumany.get_bunrui(Hinsi, Bunrui), jumany.get_katuyou1(Katuyou1),
            jumany.get_katuyou2(Katuyou1, Katuyou2)
        ]
        if args.delimiter is None:
            column_print(mrph)
        else:
            sys.stdout.write(args.delimiter.join(mrph)+"\n")
    print("EOS")



"""
jumany
Python interface for JUMAN
"""
from .juman_ext import SUCCESS, MRPH_BUFF_OVER, WORD_BUFF_OVER
from .juman_ext import ERROR, RC_ERROR, MALLOC_ERROR, FILE_NOT_EXISTS
from .juman_ext import ANA_ERROR, PYMOD_ERROR, TOO_LONG

from .juman_ext import ERRORNO, L_HINSI, L_BUNRUI, L_KATUYOU1, L_KATUYOU2

from .juman_ext import get_error_msg, status, open_lib, close_lib, analyze
from .juman_ext import get_hinsi, get_bunrui, get_katuyou1, get_katuyou2
# from .juman_ext import set_encoding



"""
    juman ext api
"""
import os
import sys
import glob
from ctypes import CDLL, Structure, POINTER, byref, cast
from ctypes import c_int, c_char_p, c_size_t, c_byte, c_void_p
import re

def load_lib():
    """ Load shared library """
    libs = glob.glob(os.path.dirname(__file__) + '/libjuman.*so')
    if not libs:
        libs = glob.glob(os.path.dirname(__file__) + '/../dist/juman-*/lib/*.so')
        if not libs:
            print("Cannot find libjuman.so")
            sys.exit(-1)
    if len(libs) > 1: # To narrow down candidates
        if os.name == "nt":
            platform = "win64" if sys.maxsize > 2**32 else "win32"
            pl_libs = filter(lambda x: x.find(platform) >= 0, libs)
        else:   # POSIX
            pl_libs = filter(lambda x: x.find("win64") < 0 and x.find("win32") < 0, libs)
        if pl_libs:
            libs = list(pl_libs)
    return CDLL(libs[0])

LIBC = load_lib()

class MrphT(Structure):  # pylint: disable=R0903
    """ Type of morphese """
    _fields_ = [
        ("midasi1", c_char_p),
        ("yomi", c_char_p),
        ("midasi2", c_char_p),
        ("hinsi", c_int),
        ("bunrui", c_int),
        ("katuyou1", c_int),
        ("katuyou2", c_int)
    ]

#	EXT_RES_CODE
SUCCESS = 0
MRPH_BUFF_OVER = 1
WORD_BUFF_OVER = 2
ERROR = 100
RC_ERROR = 101
MALLOC_ERROR = 102
FILE_NOT_EXISTS = 103
ANA_ERROR = 104
PYMOD_ERROR = 200
TOO_LONG = 201

# void ext_set_encoding(const char *s_encoding);
LIBC.ext_init.restype = None
LIBC.ext_init.argtypes = (c_char_p,)

# EXT_RES_CODE ext_init(const char *s_rcfile, size_t word_buf_size, int max_mrphs);
LIBC.ext_init.restype = c_int
LIBC.ext_init.argtypes = (c_char_p, c_size_t, c_int)

# void ext_close();
LIBC.ext_close.restype = None
LIBC.ext_close.argtypes = ()

# char * ext_get_input_buff(size_t *psize);
LIBC.ext_get_input_buff.restype = POINTER(c_byte)
LIBC.ext_get_input_buff.argtypes = (POINTER(c_size_t),)

# EXT_MRPH_T * ext_get_mrph_buff();
LIBC.ext_get_mrph_buff.restype = POINTER(MrphT)
LIBC.ext_get_mrph_buff.argtypes = ()

# EXT_RES_CODE ext_analyze();
LIBC.ext_analyze.restype = c_int
LIBC.ext_analyze.argtypes = ()

# EXT_RES_CODE ext_get_result(int *pnum_mrphs);
LIBC.ext_get_result.restype = c_int
LIBC.ext_get_result.argtypes = (POINTER(c_int),)

# const char * ext_get_hinsi(int hinsi);
LIBC.ext_get_hinsi.restype = c_char_p
LIBC.ext_get_hinsi.argtypes = (c_int,)

# int ext_get_all_hinsi(const char ** list);
LIBC.ext_get_all_hinsi.restype = c_int
LIBC.ext_get_all_hinsi.argtypes = (POINTER(c_char_p),)

# const char * ext_get_bunrui(int hinsi, int bunrui);
LIBC.ext_get_bunrui.restype = c_char_p
LIBC.ext_get_bunrui.argtypes = (c_int, c_int)

# int ext_get_all_bunrui(int hinsi, const char ** list);
LIBC.ext_get_all_bunrui.restype = c_int
LIBC.ext_get_all_bunrui.argtypes = (c_int, POINTER(c_char_p))

# const char * ext_get_katuyou1(int katuyou1);
LIBC.ext_get_katuyou1.restype = c_char_p
LIBC.ext_get_katuyou1.argtypes = (c_int,)

# int ext_get_all_katuyou1(const char ** list);
LIBC.ext_get_all_katuyou1.restype = c_int
LIBC.ext_get_all_katuyou1.argtypes = (POINTER(c_char_p),)

# const char * ext_get_katuyou2(int katuyou1, int katuyou2);
LIBC.ext_get_katuyou2.restype = c_char_p
LIBC.ext_get_katuyou2.argtypes = (c_int, c_int)

# int ext_get_all_katuyou2(int katuyou1, const char ** list);
LIBC.ext_get_all_katuyou2.restype = c_int
LIBC.ext_get_all_katuyou2.argtypes = (c_int, POINTER(c_char_p))

# void ext_get_maxidx(int *pmax_hinsi, int *pmax_bunrui,
#	int *pmax_katuyou1, int *pmax_hatuyou2);
LIBC.ext_get_maxidx.restype = None
LIBC.ext_get_maxidx.argtypes = (POINTER(c_int), POINTER(c_int), POINTER(c_int), POINTER(c_int))

ERRORNO = SUCCESS

L_HINSI = []
L_BUNRUI = []
L_KATUYOU1 = []
L_KATUYOU2 = []

_JUMANRC = None
_WRITE_BUF = None
_BUF_SIZE = None
_READ_BUF = None
_ENCODING = "UTF-8"

_RE_EMPTY = re.compile(r"^[ \t\r\n]+$")

def get_error_msg()->str:
    """ Get last error message
    @return Error message. Or null string when no error.
    """
    msg = "Something wrong."
    if ERRORNO == SUCCESS:
        msg = ""
    elif ERRORNO == MRPH_BUFF_OVER:
        msg = "Morpheme buffer is not enough. Need to call several times."
    elif ERRORNO == WORD_BUFF_OVER:
        msg = "Word buffer is not enough. Need to call several times."
    elif ERRORNO == ERROR:
        msg = "Error in previous call."
    elif ERRORNO == RC_ERROR:
        msg = "Error in jumanrc."
    elif ERRORNO == MALLOC_ERROR:
        msg = "Fail to allocate buffer."
    elif ERRORNO == FILE_NOT_EXISTS:
        msg = "File does not exist."
    elif ERRORNO == ANA_ERROR:
        msg = "Error in analyze."
    elif ERRORNO == TOO_LONG:
        msg = "Input text is too long."
    return msg

def status(stream=None):
    """ Show status
    @param stream Stream to output
    @return String of status
    """
    stat_str = ""
    if _JUMANRC is None:
        stat_str += "Closed."
    else:
        stat_str += "Resource file: %s\n" % _JUMANRC
        ptr = cast(_WRITE_BUF, c_void_p)
        stat_str += "Write buffer : %s\n" % hex(ptr.value)
        stat_str += "Buffer size  : %d\n" % _BUF_SIZE
        ptr = cast(_READ_BUF, c_void_p)
        stat_str += "Read buffer  : %s\n" % hex(ptr.value)
        stat_str += "ERRORNO      : %d\n" % ERRORNO
        stat_str += "message : %s\n" % get_error_msg()
    if stream:
        stream.write(stat_str)
    return stat_str

def open_lib(rc_path: str = None)->bool:
    """ Open juman library
    @param rc_path Path to resource file. None is available to load default.
    @return True when success, otherwise False.
    """
    global _JUMANRC, ERRORNO, _WRITE_BUF, _BUF_SIZE, _READ_BUF
    global L_HINSI, L_BUNRUI, L_KATUYOU1, L_KATUYOU2
    ERRORNO = SUCCESS
    if _JUMANRC is not None:
        return True
    if rc_path is None:
        abspath = os.path.abspath(__file__)
        rc_path = os.path.join(os.path.dirname(abspath), 'jumanrc')
    if not os.path.exists(rc_path):
        ERRORNO = FILE_NOT_EXISTS
        return False
    res = LIBC.ext_init(rc_path.encode('UTF-8'), 0, 0)
    if SUCCESS != res:
        ERRORNO = res
        return False
    _JUMANRC = rc_path
    size = c_size_t()
    _WRITE_BUF = LIBC.ext_get_input_buff(byref(size))
    _BUF_SIZE = size.value
    _WRITE_BUF = cast(_WRITE_BUF, POINTER(c_byte * _BUF_SIZE))
    _READ_BUF = LIBC.ext_get_mrph_buff()
    # Get all word of grammar
    max_param = [c_int() for i in range(4)]
    LIBC.ext_get_maxidx(
        byref(max_param[0]), byref(max_param[1]),
        byref(max_param[2]), byref(max_param[3]))
    max_param = (x.value for x in max_param)
    max_num = max(max_param)
    alist = (c_char_p * max_num)()
    hinsi_num = LIBC.ext_get_all_hinsi(alist)
    for i in range(0, hinsi_num):
        L_HINSI.append(alist[i].decode("UTF-8"))
    for i in range(0, hinsi_num):
        sub_num = LIBC.ext_get_all_bunrui(i, alist)
        sublist = []
        for j in range(0, sub_num):
            sublist.append(alist[j].decode("UTF-8"))
        L_BUNRUI.append(sublist)
    k1_num = LIBC.ext_get_all_katuyou1(alist)
    for i in range(0, k1_num):
        L_KATUYOU1.append(alist[i].decode("UTF-8"))
    for i in range(0, k1_num):
        sub_num = LIBC.ext_get_all_katuyou2(i, alist)
        sublist = []
        for j in range(0, sub_num):
            sublist.append(alist[j].decode("UTF-8"))
        L_KATUYOU2.append(sublist)

    return True

def close_lib():
    """ Close juman library  """
    global _JUMANRC, ERRORNO, _WRITE_BUF, _BUF_SIZE, _READ_BUF
    global L_HINSI, L_BUNRUI, L_KATUYOU1, L_KATUYOU2
    if _JUMANRC is None:
        return
    LIBC.ext_close()
    # Clear
    _JUMANRC = None
    _WRITE_BUF = None
    _BUF_SIZE = None
    _READ_BUF = None
    _ENCODING = "UTF-8"
    L_HINSI = []
    L_BUNRUI = []
    L_KATUYOU1 = []
    L_KATUYOU2 = []

def set_encoding(encoding: str):
    """ Set encoding of input text
    @param encoding String of encoding
    @todo need to consider working with bytes data.
    """
    global _ENCODING
    _ENCODING = encoding

def analyze(text: str, remove_space: bool = False, just_word: bool = False)->[]:
    """ Analyze sentences
    @param text String to analyze
    @param remove_space If True, remove morpheme consist from only space, CR, LF, and tab.
    @param just_word If True, only word in text is set in list.
    Otherwise tuple of params is set.
    @return List of morpheme. If None, Error occured.
    If just_word is True, list likes [word:str, word:str, ... ].
    If just_word is False, list likes
    [ (midasi1:str, yimi:str, midasi2:str, hinsi:int,
        bunrui:int, katuyou1:int, katutou2:int), ... ]
    """
    global ERRORNO, _ENCODING
    if _do_analyze(text) is False:
        return None
    # Get result
    num = c_int()
    mrphs = []
    res = -1
    while res != SUCCESS:
        res = LIBC.ext_get_result(byref(num))
        if res >= ERROR:
            ERRORNO = res
            return None
        for i in range(0, num.value):
            word = _READ_BUF[i].midasi1.decode("UTF-8")
            if remove_space and re.match(_RE_EMPTY, word) is not None:
                continue
            if just_word:
                mrphs.append(word)
            else:
                mrph = _READ_BUF[i]
                mrphs.append((
                    word, mrph.yomi.decode("UTF-8"),
                    mrph.midasi2.decode("UTF-8"), mrph.hinsi,
                    mrph.bunrui, mrph.katuyou1, mrph.katuyou2
                ))
    return mrphs

def _do_analyze(text: str)->bool:
    """ Execute analyze part
    @param text String to analyze
    @return True when success, otherwise False.
    """
    global ERRORNO, _ENCODING
    data = text.encode(_ENCODING)
    data_len = len(data)
    if data_len >= _BUF_SIZE:
        ERRORNO = TOO_LONG
        return False
    _WRITE_BUF.contents[:data_len] = data[:]
    _WRITE_BUF.contents[data_len] = 0
    res = LIBC.ext_analyze()
    if res != SUCCESS:
        ERRORNO = res
        return False
    return True

def get_hinsi(hinsi: int)->str:
    """ Get honsi from code
    @param hinsi Hinsi code
    @reurn String of hinsi
    """
    global L_HINSI
    return L_HINSI[hinsi]

def get_bunrui(hinsi: int, bunrui: int)->str:
    """ Get bunrui from codes
    @param hinsi Hinsi code
    @param bunrui bunrui code
    @reurn String of bunrui
    """
    global L_BUNRUI
    return L_BUNRUI[hinsi][bunrui]

def get_katuyou1(katuyou1: int)->str:
    """ Get katuyou1 from code
    @param katuyou1 Katuyou1 code
    @reurn String of katuyou1
    """
    global L_KATUYOU1
    return L_KATUYOU1[katuyou1]

def get_katuyou2(katuyou1: int, katuyou2: int)->str:
    """ Get katuyou2 from codes
    @param katuyou1 Katuyou1 code
    @param katuyou2 Katuyou2 code
    @reurn String of katuyou2
    """
    global L_KATUYOU2
    return L_KATUYOU2[katuyou1][katuyou2]

if __name__ == '__main__':
    from test import get_jumanrc
    status(sys.stderr)
    if open_lib(get_jumanrc()) is False:
        print(get_error_msg())
        sys.exit(0)
    status(sys.stderr)
    Text = """吾輩は猫である。名前はまだ無い。\t
どこで生れたかとんと見当がつかぬ。何でも薄暗いじめじめした所でニャーニャー泣いていた事だけは記憶している。吾輩はここで始めて人間というものを見た。"""
    Mrphs = analyze(Text)
    if Mrphs is None:
        print(get_error_msg())
        sys.exit(0)
    print(repr(Mrphs))
    for (Midasi1, Yomi, Midasi2, Hinsi, Bunrui, Katuyou1, Katuyou2) in Mrphs:
        print("{} : {} : {} : {} : {} : {} : {}".format(
            Midasi1, Yomi, Midasi2, get_hinsi(Hinsi),
            get_bunrui(Hinsi, Bunrui), get_katuyou1(Katuyou1),
            get_katuyou2(Katuyou1, Katuyou2)
        ))
    print(analyze(Text, True, True))
    close_lib()



"""
jumany setup
"""
from scripts.setup_ext import SetupExt

# Don't insert space to be compatible shell script.
MODULE_NAME="jumany" # pylint: disable=C0326
MODULE_VERSION="0.4.1" # pylint: disable=C0326

DEVELOPMENT_STATE = 4   # Beta
PACKAGE = "jumany"

_EXT = SetupExt(
    name=MODULE_NAME,
    version=MODULE_VERSION,
    description='Interface for JUMAN Morphological analysis system',
    packages=[PACKAGE, PACKAGE+'.test'],
    package_dir={
        PACKAGE: 'python_module',
        PACKAGE+'.test': 'python_module/test'
    },
    package_data={PACKAGE: ['README.*', '*.txt']},
    # test_suite='jumany.test',
    author='yujakudo',
    url='https://github.com/yujakudo/jumany',
    classifiers=[
        # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Topic :: Text Processing",
    ],
)

_EXT.set_requirement(
    python=[(3, 3)],
    win64=True,
    win32=True,
    posix=True,
)

_EXT.set_development_state(DEVELOPMENT_STATE)

_EXT.set_extensions(
    posix={
        'libjuman': dict(
            description='library of JUMAN',
            src_dir='juman-7.01/lib',
            define_macros=[('HAVE_CONFIG_H', 1)],
            include_dirs=['juman-7.01/lib', 'scripts'],
            extra_compile_args=["-Wno-error=format-security", "-O3", "-march=native"],
        )
    }
)

_EXT.set_custom_build(
    for_all=[
        'copy scripts/jumanrc.win jumanrc',
        'copy_tree dist/juman-7.01_ext_win64/dics dics'
    ],
    win64=[
        'copy dist/juman-7.01_ext_win64/lib/libjuman.so libjuman.win64.so',
        'copy copyings/copying.libjuman.win.txt LICENSE.libjuman.txt',
    ],
    win32=[
        'copy dist/juman-7.01_ext_win32/lib/libjuman.so libjuman.win32.so',
        'copy copyings/copying.libjuman.win.txt LICENSE.libjuman.txt',
    ],
    posix=[
        'copy copyings/copying.libjuman.txt LICENSE.libjuman.txt',
    ],
)

_EXT.setup()



"""
Module tests
"""
import os
import sys
import unittest
sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
try:
    import jumany
except ImportError:
    import python_module as jumany

def get_jumanrc():
    """ Get testable jumanrc """
    mod_dir = os.path.join(os.path.dirname(__file__), "..")
    candidates = [
        os.path.join(mod_dir, "jumanrc"),
        os.path.join(mod_dir, "../dist/juman-7.01_ext_win64/jumanrc"),
    ]
    for rc in candidates:
        if os.path.exists(rc):
            return rc
    return None

_TEST_STR = "吾輩は猫である。"
_TEST_EXPECT = [
    ("吾輩", "名詞"), ("は", "助詞"), ("猫", "名詞"), ("である", "判定詞"), ("。", "特殊")
]

class test_jumany_open(unittest.TestCase):
    """ test open and close"""
    def test_open_fail(self):
        """  Fail to jumany.open_lib()    """
        jumany.close_lib()
        self.assertFalse(jumany.open_lib("hoge"))
        msg = jumany.get_error_msg()
        self.assertTrue(len(msg) > 0)

    def test_open_success(self):
        """  Success to jumany.open_lib()    """
        jumanrc = get_jumanrc()
        self.assertTrue(jumany.open_lib(jumanrc))
        msg = jumany.get_error_msg()
        self.assertTrue(len(msg) == 0)
        jumany.close_lib()

class test_jumany_main(unittest.TestCase):
    """ For jumany main"""

    def setUp(self):
        jumanrc = get_jumanrc()
        jumany.open_lib(jumanrc)

    def tearDown(self):
        #jumany.close_lib()
        pass

    def test_analyze_normal(self):
        """  jumany.analyze() in nomal use  """
        res = jumany.analyze(_TEST_STR)
        self.assertEqual(len(_TEST_EXPECT), len(res))
        self.assertEqual(7, len(res[0]))
        for expect, result in zip(_TEST_EXPECT, res):
            self.assertEqual(expect[0], result[0])
            self.assertEqual(expect[1], jumany.get_hinsi(result[3]))

    def test_analyze_just_word(self):
        """  jumany.analyze() for just_word==True  """
        res = jumany.analyze(_TEST_STR, just_word=True)
        self.assertEqual(len(_TEST_EXPECT), len(res))
        for expect, result in zip(_TEST_EXPECT, res):
            self.assertEqual(expect[0], result)

    def test_analyze_remove_space1(self):
        """  jumany.analyze() for remove_space==False  """
        res = jumany.analyze(" \r\t\n", remove_space=False)
        self.assertEqual(2, len(res))

    def test_analyze_remove_space2(self):
        """  jumany.analyze() for remove_space==True  """
        res = jumany.analyze(" \r\t\n", remove_space=True)
        self.assertEqual(0, len(res))

if __name__ == "__main__":
    unittest.main()



"""
setup extention
"""
import os
import sys
import glob
import shlex
import inspect
import pprint
from distutils.command.build import build
import setuptools

_SYSTEMS = {
    'posix': 'POSIX',
    'win32': 'Windows 32bit',
    'win64': 'Windows 64bit',
}

_SYSTEM_CLASSFIERS = {
    'posix': "Operating System :: POSIX",
    'win32': "Operating System :: Microsoft :: Windows",
    'win64': "Operating System :: Microsoft :: Windows",
}

_DEV_STATUS_CLASSFIERS = [
    '',
    "Development Status :: 1 - Planning",
    "Development Status :: 2 - Pre-Alpha",
    "Development Status :: 3 - Alpha",
    "Development Status :: 4 - Beta",
    "Development Status :: 5 - Production/Stable",
    "Development Status :: 6 - Mature",
    "Development Status :: 7 - Inactive",
]

_PY_MAX_MINORS = [0, 0, 7, 6, 0]

class SetupExt: # pylint: disable=R0902
    """ Setup extention """

    def __init__(self, **setup_args):
        """ condtructor
        @param setup_args arguments of setup to extent
        """
        self.args = setup_args
        self.system_req = None
        self.python_req_min = [None for x in _PY_MAX_MINORS]
        self.caller = inspect.currentframe().f_back.f_code.co_filename
        self.base_dir = os.path.dirname(self.caller)
        # Get system
        self.system = None
        if os.name == 'nt':
            if sys.maxsize <= 2**32:
                self.system = 'win32'
            else:
                self.system = 'win64'
        elif os.name == 'posix':
            self.system = 'posix'
        # Get package name and directory
        self.package_name = self.args['packages'][0]
        self.package_dir = os.path.join(self.base_dir, self.package_name)
        if 'package_dir' in self.args and self.package_name in self.args['package_dir']:
            self.package_dir = os.path.join(
                self.base_dir, self.args['package_dir'][self.package_name]
            )
        # Set long descrition
        if 'long_description' not in self.args:
            self.args['long_description'] = self.args['description']
            fn = None
            for ext in ['', '.rst', '.txt']:
                fn = os.path.join(self.package_dir, 'README'+ext)
                if os.path.exists(fn):
                    with open(fn, 'rb') as fp:
                        self.args['long_description'] = fp.read().decode('utf8')
                    break

    def set_development_state(self, code: int):
        """ Set development state
        @param code development state from 1 to 6.(3 to 5 are common)
        """
        if 'classifiers' not in self.args:
            self.args['classifiers'] = []
        self.args['classifiers'].append(_DEV_STATUS_CLASSFIERS[code])

    def set_requirement(self, **req):
        """ Set system requirement
        @param python List of taple of minimum version to allow.
        e.g.[(2,5), (3,3)]
        @param posix, win32, win64 Not False if supported
        """
        self.system_req = req
        if 'classifiers' not in self.args:
            self.args['classifiers'] = []
        for sys_name in _SYSTEMS:
            if sys_name in self.system_req and self.system_req[sys_name] is not False:
                classfier = _SYSTEM_CLASSFIERS[sys_name]
                if classfier not in self.args['classifiers']:
                    self.args['classifiers'].append(classfier)
        if 'python' in self.system_req:
            max_major = 0
            pyreq = ''
            for ver in self.system_req['python']:
                self.python_req_min[ver[0]] = ver[1]
                pyreq += ', ' if pyreq != '' else ''
                pyreq += ">=%d.%d" % (ver[0], ver[1])
                max_major = ver[0] if ver[0] > max_major else max_major
                max_minor = _PY_MAX_MINORS[ver[0]]
                for minor in range(ver[1], max_minor + 1):
                    self.args['classifiers'].append(
                        "Programming Language :: Python :: %d.%d" % (ver[0], minor)
                    )
            if max_major > 0:
                if 'python_requires' not in self.args:
                    self.args['python_requires'] = pyreq + ", <%d" % (max_major + 1)

    def set_extensions(self, **exts):
        """ Set info. to build extension libraries
        Each parameter takes a dict of extension.
        Value is saame as setuptools.Extension(), but added parameters.
        - description: Short description for error or help.
        - src_dir: Source directory to extract to 'sources'
        @param posix, win32, win64
        """
        if self.system not in exts:
            return
        features = {}
        for name, ext_args in exts[self.system].items():
            if not 'name' in ext_args:
                ext_args['name'] = self.package_name + '.' + name
            if 'src_dir' in ext_args:
                ext_args['sources'] = glob.glob(ext_args['src_dir'] + '/*.c*')
                del ext_args['src_dir']
            desc = ext_args['name']
            if 'description' in ext_args:
                desc = ext_args['description']
                del ext_args['description']
            features[name] = setuptools.Feature(
                desc, standard=True,
                ext_modules=[setuptools.Extension(**ext_args)]
            )
        self.args['features'] = features

    def set_custom_build(self, for_all=None, **custom):
        """ Custom build
        Each parameter is a List of commandline.
        Parameter dest is relative path from package directory.
        Commands are follows:
        - copy: 'copy src dest': Copy file src to dest.
        - copy_tree: 'copy_tree src dest': Copy directory src to dest.
        - move: 'move src dest': Move file src to dest.
        - mkpath: 'mkpath dest mode': Make directory dest. default of mode is '0o777'
        - execute: 'execute func args...': ececute function
        - spawn: 'spawn cmd': execute cmd in subprocess.
        @param for_all, posix, win64, win32
        """
        CustomBuild.set_pakage_name(self.package_name)
        if isinstance(for_all, list):
            CustomBuild.set_rules(for_all)
        if self.system in custom:
            CustomBuild.set_rules(custom[self.system])
        if CustomBuild.count_rules() == 0:
            return
        if 'cmdclass' not in self.args:
            self.args['cmdclass'] = {}
        self.args['cmdclass']['build'] = CustomBuild

    def get_args(self)->dict:
        """ Get setup arguments """
        return self.args

    def _show(self):
        """ Show arguments and exit when 'show' entered """
        if "show" not in sys.argv:
            return
        pp = pprint.PrettyPrinter(indent=2)
        pp.pprint(self.args)
        sys.exit(0)

    def _check_requirement(self):
        """ Check requirement when install
        Exit if this system not supported.
        """
        if "install" not in sys.argv:
            return
        msg = ""
        if not self.system in self.system_req or self.system_req[self.system] is False:
            msg += "%s system is not supported.\n" % _SYSTEMS[self.system]
        ver = sys.version_info
        min_minor = self.python_req_min[ver[0]]
        if min_minor is None or ver[1] < min_minor:
            msg += "Python %d.%d is not supported.\n" % (ver[0], ver[1])
            msg += "Followings are required.\n"
            for majour, min_minor in enumerate(self.python_req_min):
                if majour:
                    msg += "- Python %d.%d or above.\n" % (majour, min_minor)
        if msg != "":
            sys.stderr.write(
                "%s can not be installed on this system.\n" % self.package_name
            )
            sys.stderr.write(msg)
            sys.exit(-1)

    def setup(self):
        """ Do setup """
        self._show()
        self._check_requirement()
        setuptools.setup(**self.args)

class CustomBuild(build):
    """ Overwrite build """
    _rules = []
    _package_name = None
    @staticmethod
    def set_pakage_name(name: list):
        """ set package name """
        CustomBuild._package_name = name
    @staticmethod
    def set_rules(rules: list):
        """ set rule """
        CustomBuild._rules += rules
    @staticmethod
    def count_rules():
        """ Get number of rules """
        return len(CustomBuild._rules)

    def path_from_package(self, path: str)->str:
        """ get path from package directory """
        return os.path.join(self.build_lib, CustomBuild._package_name, path)

    def run(self):
        """ run build process """
        build.run(self)
        for rule in CustomBuild._rules:
            args = shlex.split(rule)
            cmd = args.pop(0)
            if cmd == 'copy':
                args[1] = self.path_from_package(args[1])
                self.copy_file(*args)
            elif cmd == 'copy_tree':
                args[1] = self.path_from_package(args[1])
                self.copy_tree(*args)
            elif cmd == 'move':
                args[1] = self.path_from_package(args[1])
                self.move_file(*args)
            elif cmd == 'mkpath':
                args[0] = self.path_from_package(args[0])
                self.mkpath(*args)
            elif cmd == 'execute':
                func = args.pop(0)
                self.execute(func, args)
            elif cmd == 'spawn':
                self.spawn(*args)

if __name__ == '__main__':
    _EXT = SetupExt(
        name="test_module",
        version="0.1",
        description='Test module',
        packages=['jumany'],
        package_dir={'jumany': 'python_module'},
    )

    _EXT.set_requirement(
        python=[(2, 5), (3, 3)],
        win64=True,
        win32=True,
        posix=True,
    )
    _EXT.set_development_state(4)
    _EXT.set_extensions(
        win64={
            'libjuman': dict(
                description='library of JUMAN',
                src_dir='juman-7.01/lib',
            )
        }
    )
    _EXT.set_custom_build(
        win64=[
            'copy dist/juman-7.01_ext_win64/lib/libjuman.so libjuman.win64.so',
        ],
    )
    sys.argv = ["hoge", "install"]
    print(repr(_EXT.get_args()))



"""
Definision of test suite
"""
import os
from unittest import TestLoader, TextTestRunner

def run():
    """ test suite """
    tests = TestLoader().discover(os.path.dirname(__file__))
    TextTestRunner(verbosity=2).run(tests)

if __name__ == '__main__':
    run()
